#include <dx/dx.h>
#include <math.h>
  
/* two quick macros to make landscape generation easier */
#define F3(delta,x0,x1,x2)      ((x0+x1+x2)/3 + delta*Gauss())
#define F4(delta,x0,x1,x2,x3)   ((x0+x1+x2+x3)/4 + delta*Gauss())


/*
   ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   |                                                            |
   |            Global Variables                                |
   |                                                            |
   ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/

int     Nrand;          /* Number of samples of random to use */
long    Arand;          /* Maximum value of random number */
double  GaussAdd,
        GaussFac;       /* a real parameter for linear transform in gauss */

double  **X;		/* 2d array to hold landscape heights */
double  *X1;		/* 1d array to hold landscape hieghts */
double  numer, denom;
int	size;

  
static Error DoMakeLandscape(Object in, float h_param, 
			float seed_param, float sigma_param);


m_Landscape(Object *in, Object *out)
{
  Object o = NULL;
  float h_param, seed_param, sigma_param;

  if (!in[0]) {
    DXSetError(ERROR_BAD_PARAMETER, "missing object");
    goto error;
  }
  o = (Object)DXCopy(in[0], COPY_STRUCTURE);
  if (!o)
    goto error;

  /* DXExtract floating point parameter from in[1] (default 1) */
  if (!in[1])
    h_param = 1;
  else if (!DXExtractFloat(in[1], &h_param)) {
    DXSetError(ERROR_BAD_PARAMETER, "H must be a scalar");
    goto error;
  }

  if (!in[2])
    sigma_param = 1;
  else if (!DXExtractFloat(in[2], &sigma_param)) {
    DXSetError(ERROR_BAD_PARAMETER, "Sigma must be a scalar");
    goto error;
  }

  if (!in[3])
    seed_param = 1;
  else if (!DXExtractFloat(in[3], &seed_param)) {
    DXSetError(ERROR_BAD_PARAMETER, "Seed must be a scalar");
    goto error;
  }


  /*	Make the landscape */
  if (!DoMakeLandscape(o, h_param, seed_param, sigma_param))
    goto error;
  
  /* successful return */
  out[0] = o;
  return OK;
  
 error:
  DXDelete(o);
  return ERROR;
}


InitGauss(seed)
int     seed;
{
    Nrand = 4;
    Arand = (1<<31)-1;          /* random() gives results in [0,Arand] */

    GaussAdd = (double)sqrt((double)(3.0*Nrand));
    numer = GaussAdd + GaussAdd;
    denom = (double)((double)Nrand * (double)Arand);

    srandom(seed);
}

double
Gauss()
{
register int    i;
register double sum = 0;
extern long     random();

    for (i=0; i<Nrand; i++)
        sum += (double)random();

    sum = numer*(sum/denom);
    return(sum - GaussAdd);
}


MidPointFM2D(Xo,maxlevel,sigma,H,addition,seed)
double **Xo;
int     maxlevel;
double  sigma;
double  H;
char    addition;
int     seed;
{
    register int    i, N, x,y, D,d;
    double          delta;
	
    InitGauss(seed);
    N = pow(2,maxlevel);
    delta = sigma;

    Xo[0][0] = delta * Gauss();
    Xo[0][N] = delta * Gauss();
    Xo[N][0] = delta * Gauss();
    Xo[N][N] = delta * Gauss();

    D = N;
    d = N/2;

    for (i=0; i<maxlevel; i++) {

        delta = delta * pow((double)0.5,(double)(0.5*H));

        for (x=d; x<=N-d; x+=D)       
            for (y=d; y<=N-d; y+=D)
                Xo[x][y] = F4(delta,Xo[x+d][y+d],Xo[x+d][y-d],
					Xo[x-d][y+d],Xo[x-d][x-d]);

        if (addition == 1)
            for (x=0; x<=N; x+=D)
                for (y=0; y<=N; y+=D)
                    Xo[x][y] = Xo[x][y] + delta*Gauss();

        delta = delta * pow(0.5,0.5*H);

        for (x=d; x<=N-d; x+=D) {
            Xo[x][0] = F3(delta,Xo[x+d][0],Xo[x-d][0],Xo[x][d]);
            Xo[x][N] = F3(delta,Xo[x+d][N],Xo[x-d][N],Xo[x][N-d]);
            Xo[0][x] = F3(delta,Xo[0][x+d],Xo[0][x-d],Xo[d][x]);
            Xo[N][x] = F3(delta,Xo[N][x+d],Xo[N][x-d],Xo[N-d][x]);
        }

        for (x=d; x<=N-d; x+=D)
            for (y=D; y<=N-d; y+=D)
                Xo[x][y] = F4(delta,Xo[x][y+d],Xo[x][y-d],
				Xo[x+d][y],Xo[x-d][y]);

        for (x=D; x<=N-d; x+=D)
            for (y=d; y<=N-d; y+=D)
                Xo[x][y] = F4(delta,Xo[x][y+d],Xo[x][y-d],
				Xo[x+d][y],Xo[x-d][y]);

        if (addition == 1) {
            for (x=0; x<=N; x+=D)
                for (y=0; y<=N; y+=D)
                    Xo[x][y] = Xo[x][y] + delta*Gauss();

            for (x=d; x<=N-d; x+=D)
                for (y=d; y<=N-d; y+=D)
                    Xo[x][y] = Xo[x][y] + delta*Gauss();
        }

        D = D/2;
        d = d/2;
    }
}



static
  Error
  DoMakeLandscape(Object o, float h_param, float seed_param, float sigma_param)
{
  Array old_positions=NULL, new_positions=NULL, 
	new_connections=NULL,a=NULL;
  Object oo;
  float *old_pos, *new_pos; 
  int *new_con;
  int i, n, j, count;
  
  /* determine the class of the object */
  switch (DXGetObjectClass(o)) {
    
  case CLASS_FIELD:
    
    /* 
     * DXExtract the ``positions'' component array from the input
     * object.  Then typecheck it, determine its size, and get
     * a pointer to the array data.
     */
    old_positions = (Array) DXGetComponentValue((Field)o, "positions");

    if (!old_positions) {
      DXSetError(ERROR_MISSING_DATA, "input has no positions");
      return ERROR;
    }

    if (!DXTypeCheck(old_positions, TYPE_FLOAT, CATEGORY_REAL, 1, 3)) {
      DXSetError(ERROR_BAD_TYPE, 
		 "positions are not 3D floating point");
      return ERROR;
    }

    /*
     * Use GetArrayInfo to get size of array
     */

    if (!DXGetArrayInfo(old_positions, &n, NULL, NULL, NULL, NULL))
      return ERROR;

    /*
     * Extract the position values w/ get array data
    */

    old_pos = (float *) DXGetArrayData(old_positions);
    if (!old_pos)
      return ERROR;
    
    /* create a new array, allocate space in it, put it in the field */

    a = (Array)DXNewArray(TYPE_FLOAT, CATEGORY_REAL, 1, 3);

    if (!DXAddArrayData(a, 0, n, NULL))
      return ERROR;

    new_pos = (float *) DXGetArrayData(a);

    DXSetComponentValue((Field)o, "positions", (Object)a);


    /*
     * Here's where the meat begins. First find x and y dim 
     * Then alloc space in a 2d array of x by y. 
     */

    size = (int) sqrt(n);
    X = calloc( size+1, size * sizeof(double*));

    for (i = 0; i < size+1; i++)
	X[i] = malloc( size * sizeof(double));

    /*
     * Call MidPointFM2D (generate the landscape).
     * Use the 2d array and the user's parameters.
     * Since we have width, and MidPoint needs maxlevel
     * 2^maxlevel = width so log(base2) of size = maxlevel.
     */

   MidPointFM2D(X, (int) (log(size)/log(2)), sigma_param, 
	h_param, TRUE, seed_param);
 
    /*
     * Squeze the 2d array into a 1d array of size width^2.
     */

   count = 0;
	X1 = (double*) malloc(n*sizeof(double));
 	for (i = 0; i < size; i++)
	{
	  for (j = 0; j < size; j++)
	  {
		X1[count] = X[i][j];
		count++;
	  }
	}

    /*
     * Loop for the total number of positions.
     * Set the x to the x, z to the z, y to the landscape value.
     */

  count = 0;
	for (i = 0; i < n; i++){
      		new_pos[3*count] = old_pos[3*count];
      		new_pos[3*count+1] = X1[count];
      		new_pos[3*count+2] = old_pos[3*count+2];
		count ++;
	}
    /*
     * Tell DX that we changed the value of positions.
     */

    DXChangedComponentValues((Field)o, "positions");
    
    /* finalize field */
    if (!DXEndField((Field)o))
      return ERROR;
    break;
    
  case CLASS_GROUP:
    /*
     * Not sure what exactly this does . . . but I'll leave it.
     */

    /* recursively traverse groups */
    for (i=0; oo=(Object)DXGetEnumeratedMember((Group)o, i, NULL); i++)
      if (!DoMakeLandscape(o, h_param, seed_param, sigma_param))
	return ERROR;
    break;

  default:
    DXSetError(ERROR_BAD_PARAMETER, "object must be a group or field");
    return ERROR;
    
  }
  
  /* successful return */
  return OK;

  /* return on error */
error:
  DXDelete((Object)new_positions);
  DXDelete((Object)new_connections);
}



